import json
import base64
import decimal

try:
    # In python 2 there is a base class for the string types that
    # we can check for. It was removed in python 3 so it will cause
    # a name error.
    _ANY_STRING = (basestring, bytes)
except NameError:
    # In python 3 string and bytes are different so we explicitly check
    # for both.
    _ANY_STRING = (str, bytes)


def handle_decimals(obj):
    # Lambda will automatically serialize decimals so we need
    # to support that as well.
    if isinstance(obj, decimal.Decimal):
        return float(obj)
    return obj


class Response(object):
    def __init__(self, data, status_code=200, headers=None):
        self.data = data
        self.status_code = status_code
        self.headers = headers or {
            'Content-Type': 'application/json',
            'Access-Control-Allow-Origin': '*'
        }

    def to_dict(self, binary_types=None):
        data = self.data
        if not isinstance(data, _ANY_STRING):
            data = json.dumps(data, default=handle_decimals)
        response = {
            'statusCode': self.status_code,
            'data': data
        }
        return response

    def _base64encode(self, data):
        if not isinstance(data, bytes):
            raise ValueError('Expected bytes type for body with binary '
                             'Content-Type. Got %s type body instead.'
                             % type(data))
        data = base64.b64encode(data)
        return data.decode('ascii')

    def __call__(self):
        return self.to_dict()



import json
import decimal
from .request import Request
from .response import Response
from .errors import MellisugaError


class DecimalEncoder(json.JSONEncoder):
    def default(self, o):
        if isinstance(o, decimal.Decimal):
            return str(o)
        return super(DecimalEncoder, self).default(o)


class Api(object):
    supported = ['rest', 'ws']
    new_request = Request
    new_response = Response

    def __init__(self, event, context):
        self.event = event
        self.context = context
        self.method = None
        if 'httpMethod' in event:
            self.api = 'rest'
            self.method = event['httpMethod']
        else:
            self.api = 'ws'
        self.ws_id = None

    def _validate_headers(self, headers):
        for header, value in headers.items():
            if '\n' in value:
                raise MellisugaError("Bad value for header '%s': %r" %
                                     (header, value))

    def _request_ws(self):
        """ Will raise a KeyErro if expectet parameter are missing """
        body = json.loads(self.event['body'])
        self.ws_id = body.get('id', 0)

        return self.new_request(
            data=body.get('data', None),
            auth=self.event.get('requestContext', {}).get('authorizer', None) or {},
            params=body.get('params', None),
            route=body['action']
        )

    def _request_rest(self):
        """ Will raise a KeyErro if expectet parameter are missing """
        # TODO Check for base64 encoding

        # more than one location for parameter, merge them
        params = self.event.get('pathParameters', None) or {}
        params.update(self.event.get('queryStringParameters', None) or {})
        return self.new_request(
            data=self.event['body'],
            params=params,
            auth=self.event.get('requestContext', {}).get('authorizer', None) or {},
            headers=self.event.get('headers', None),
            route=self.event.get('path', None)
        )

    def request(self):
        return getattr(self, '_request_'+self.api)()

    def _response_ws(self, response):
        return {
            'statusCode': response.status_code,
            'body': json.dumps({
                'data': response.data,
                'id': self.ws_id,
                'status_code': response.status_code
            }, cls=DecimalEncoder)
        }

    def _response_rest(self, response):
        self._validate_headers(response.headers)
        return {
            'statusCode': response.status_code,
            'body': response.data,
            'headers': response.headers
        }

    def response(self, response):
        return getattr(self, '_response_'+self.api)(response)



import json
import base64


class Request(object):
    """The request from API gateway."""
    def __init__(self, data=None, params=None, route=None, headers=None, context=None, auth=None):
        """  Route could be path or action """
        self.params = params or {}
        self.headers = headers or {}
        self.auth = auth or {}
        self._data = data or {}
        self.route = route or ''
        self._json_data = None
        self._raw_data = b''
        self.context = context or {}
        self._is_base64_encoded = False

    def _base64decode(self, encoded):
        if not isinstance(encoded, bytes):
            encoded = encoded.encode('ascii')
        output = base64.b64decode(encoded)
        return output

    @property
    def raw_data(self):
        if not self._raw_data and self._data is not None:
            if self._is_base64_encoded:
                self._raw_data = self._base64decode(self._data)
            elif isinstance(self._data, bytes):
                self._raw_data = self._data.encode('utf-8')
            else:
                self._raw_data = self._data
        return self._raw_data

    @property
    def json_data(self):
        # TODO if self.headers.get('content-type', '').startswith('application/json'):
        if self._json_data is None:
            if isinstance(self._data, dict):
                self._json_data = self.raw_data
            else:
                try:
                    self._json_data = json.loads(self.raw_data)
                except ValueError:
                    return {}
        return self._json_data

    def to_dict(self):
        # Don't copy internal attributes.
        copied = {k: v for k, v in self.__dict__.items()
                  if not k.startswith('_')}
        # We want the output of `to_dict()` to be
        # JSON serializable, so we need to remove the CaseInsensitive dict.
        copied['headers'] = dict(copied['headers'])
        return copied




from mellisuga.mellisuga import Mellisuga
from mellisuga.response import Response
from mellisuga.request import Request
from mellisuga.api import Api
from mellisuga.errors import (
    MellisugaViewError, BadRequestError, UnauthorizedError, ForbiddenError,
    NotFoundError, ConflictError, TooManyRequestsError
)

__all__ = ["Mellisuga"]



#!/usr/bin/env python
from setuptools import setup, find_packages


with open('README.md') as readme_file:
    README = readme_file.read()


setup(
    name='mellisuga',
    version='0.0.1',
    description="AWS Lambda Microframework",
    long_description=README,
    author="Johannes Zuendel",
    author_email='jozuendel@gmail.com',
    url='https://gitlab.com/Zuendel/mellisuga',
    packages=find_packages(exclude=['tests']),
    license="Apache License 2.0",
    package_data={'mellisuga': ['*.json']},
    include_package_data=True,
    zip_safe=False,
    keywords='mellisuga aws labmda microframework websocket',
)




class MellisugaError(Exception):
    pass


class MellisugaViewError(MellisugaError):
    STATUS_CODE = 500

    def __init__(self, msg=''):
        super(MellisugaViewError, self).__init__(
            self.__class__.__name__ + ': %s' % msg)


class BadRequestError(MellisugaViewError):
    STATUS_CODE = 400


class UnauthorizedError(MellisugaViewError):
    STATUS_CODE = 401


class ForbiddenError(MellisugaViewError):
    STATUS_CODE = 403


class NotFoundError(MellisugaViewError):
    STATUS_CODE = 404


class MethodNotAllowedError(MellisugaViewError):
    STATUS_CODE = 405


class RequestTimeoutError(MellisugaViewError):
    STATUS_CODE = 408


class ConflictError(MellisugaViewError):
    STATUS_CODE = 409


class UnprocessableEntityError(MellisugaViewError):
    STATUS_CODE = 422


class TooManyRequestsError(MellisugaViewError):
    STATUS_CODE = 429


ALL_ERRORS = [
    MellisugaViewError,
    BadRequestError,
    NotFoundError,
    UnauthorizedError,
    ForbiddenError,
    RequestTimeoutError,
    ConflictError,
    UnprocessableEntityError,
    TooManyRequestsError
]



import logging
import traceback
from .api import Api
from .errors import MellisugaViewError
from functools import wraps
from collections import defaultdict


class Mellisuga(object):
    FORMAT_STRING = '%(name)s - %(levelname)s - %(message)s'

    def __init__(self, app_name, default_route=None, debug=False, api=Api):
        self.app_name = app_name
        self._debug = debug
        self.log = logging.getLogger(self.app_name)
        self.default_route = default_route
        self.actions = {}
        self.paths = defaultdict(dict)
        self.api = api

    @property
    def debug(self):
        return self._debug

    def path(self, route, **kwargs):
        def _register_path(route_func):
            def _decorator(*args, **kwargs):
                return route_func(*args, **kwargs)
            self._add_path(route, route_func, **kwargs)
            return wraps(route_func)(_decorator)
        return _register_path

    def _add_path(self, path, route_func, **kwargs):
        name = kwargs.pop('name', route_func.__name__)
        methods = kwargs.pop('methods', ['GET'])
        if kwargs:
            raise TypeError('TypeError: route() got unexpected keyword '
                            'arguments: %s' % ', '.join(list(kwargs)))
        for method in methods:
            if method in self.paths[path]:
                raise ValueError(
                    "Duplicate method: '%s' detected for route: '%s'\n"
                    "between view functions: \"%s\" and \"%s\". A specific "
                    "method may only be specified once for "
                    "a particular path." % (
                        method, path, self.paths[path][method].view_name,
                        name)
                )
            self.paths[path][method] = route_func

    def action(self, route, **kwargs):
        def _register_action(route_func):
            def _decorator(*args, **kwargs):
                # import pdb; pdb.set_trace()
                return route_func(*args, **kwargs)

            self._add_action(route, route_func, **kwargs)
            return wraps(route_func)(_decorator)
        return _register_action

    def _add_action(self, action, route_func, **kwargs):
        name = kwargs.pop('name', route_func.__name__)
        if kwargs:
            raise TypeError('TypeError: route() got unexpected keyword '
                            'arguments: %s' % ', '.join(list(kwargs)))
        if action in self.actions:
            raise ValueError("Duplicate action: '%s' detected. Name: %s" % (action, name))
        self.actions[action] = route_func

    def _error_response(self, message, error_code, http_status_code):
        body = {'Code': error_code, 'Message': message}
        response = self.api.new_response(body=body, status_code=http_status_code)
        return response

    def __call__(self, event, context):
        # This is what's invoked via lambda.
        # Sometimes the event can be something that's not
        # what we specified in our request_template mapping.
        # When that happens, we want to give a better error message here.
        try:
            api = self.api(event, context)
            request = api.request()
        except KeyError as e:
            self.log.warning("Missing parameter in Request: {}".format(e))
            return self._error_response(error_code='InternalServerError',
                                        message='Invalid api request.',
                                        http_status_code=500)

        route = request.route
        route_func = None
        if api.api == 'ws' and route in self.actions:
            route_func = self.actions[route]
        elif api.api == 'rest' and route in self.paths:
            rest_method = api.method
            if rest_method not in self.paths[route]:
                return self._error_response(
                    error_code='MethodNotAllowedError',
                    message='Unsupported method: %s' % rest_method,
                    http_status_code=405)
            route_func == self.paths[route][rest_method]
        elif self.default_route:
            route_func = self.default_route
        else:
            return self._error_response(
                error_code='MethodNotFound',
                message="No view function for: %s" % route,
                http_status_code=404)

        response = self._get_route_function_response(route_func, request, context)

        return api.response(response)

    def _get_route_function_response(self, route_function, request, context):
        try:
            response = route_function(request, context)
            if not isinstance(response, self.api.new_response):
                response = self.api.new_response(data=response)
        except MellisugaViewError as e:
            # Any mellisuga view error should propagate.  These
            # get mapped to various HTTP status codes in API Gateway.
            response = self.api.new_response(
                body={
                    'Code': e.__class__.__name__,
                    'Message': str(e)
                },
                status_code=e.STATUS_CODE
            )
        except Exception as e:
            headers = {}
            if self.debug:
                print(traceback.format_exc())
                # If the user has turned on debug mode,
                # we'll let the original exception propogate so
                # they get more information about what went wrong.
                self.log.debug("Caught exception for %s with error: %s", route_function, e,
                               exc_info=True)
                stack_trace = ''.join(traceback.format_exc())
                data = stack_trace
                headers['Content-Type'] = 'text/plain'
            else:
                data = {'Code': 'InternalServerError',
                        'Message': 'An internal server error occurred.'}
            response = self.api.new_response(data=data, headers=headers, status_code=500)
        return response
